@@ -18,7 +18,7 @@ our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
 
 our @EXPORT = qw( );
 
-our $VERSION = '0.031';
+our $VERSION = '0.05';
 
 
 ########################
@@ -56,9 +56,8 @@ sub num2cardinal {
 	$number =~s/\D+//g;
 
 	# extreme index for:
-	use constant VXN => 5;	# numbers from 5 to <1000
-	use constant THS => 6;	# тысячи
-	use constant MMT => 7;	# миллионы | миллиарды | триллионы
+	use constant VXN  => 5;	# numbers from 5 to <1000
+	use constant TMMT => 6;	# тысячи | миллионы | миллиарды | триллионы
 
 	my $gender = $cfg{'gender'} || 'masculine';
 	$gender = $gender=~/^\s*(m|f|n)/i ? @{{'m'=>'masculine', 'f'=>'feminine', 'n'=>'neuter'}}{lc $1} : 'masculine';
@@ -73,7 +72,7 @@ sub num2cardinal {
 	$object = $object=~/^\s*[a1-9]/i ? 'animate' : 'inanimate';
 
 	# Preposition (prologue) of numeral
-	my $prolog = &_ref_prolog( \%cfg );
+	my $prolog = &_ref_prolog( \%cfg, \$case );
 
 	# Epilogue of numeral
 	my $epilog = &_ref_epilog( $case, $multi, \%cfg, \$object, \$gender );
@@ -91,7 +90,12 @@ sub num2cardinal {
 		my $i = &_add_prolog( $prolog, \@s );
 
 		# Add the epilog to numeral
-		push @s, $epilog->{'root'}.$epilog->{'ends'}[0] if $epilog && exists( $epilog->{'root'} );
+		if( $epilog && exists( $epilog->{'root'} ) ) {
+			my $j = 0;
+			for my $root ( @{ $epilog->{'root'} } ) {
+				push @s, $root.$epilog->{'ends'}[ $j++ ][0];
+			}
+		}
 
 		$s[$i] = ucfirst $s[$i] if exists $cfg{'ucfirst'};
 
@@ -121,7 +125,12 @@ sub num2cardinal {
 
 			# Add epilog to numeral
 			if( $epilog && exists( $epilog->{'root'} ) ) {
-				push @s, $epilog->{'root'}.$epilog->{'ends'}[ $number < 5 ? $number : VXN ];
+
+				my $j = 0;
+				for my $root ( @{ $epilog->{'root'} } ) {
+					push @s, $root.$epilog->{'ends'}[ $j++ ][ $number < 5 ? $number : VXN ];
+				}
+
 				$epilog = {};	# epilog END
 			}
 
@@ -190,13 +199,16 @@ sub num2cardinal {
 	if( $epilog && exists( $epilog->{'root'} ) ) {
 
 		# Choice between
-		my $i = $s[-1] =~/^ты/ ? THS :	# тысяча
-				$s[-1] =~/^(?:м|трил)/ ? MMT :	# миллион | миллиард | триллион
+		my $i = $s[-1] =~/^(?:ты|ми|трил)/ ?
+				TMMT :	# тысяча | миллион | миллиард | триллион
 				VXN;	# other
 
-		push @s, $epilog->{'root'}.$epilog->{'ends'}[ $i ];
-	}
+		my $j = 0;
+		for my $root ( @{ $epilog->{'root'} } ) {
+			push @s, $root.$epilog->{'ends'}[ $j++ ][ $i ];
+		}
 
+	}
 
 	return join ' ', @s;
 }
@@ -484,7 +496,7 @@ sub _add_prolog {
 
 
 sub _ref_prolog {
-	my( $cfg ) = @_;
+	my( $cfg, $case ) = @_;
 	my $prolog = $cfg->{'prolog'} // return [ ];
 
 	if( ref($prolog) eq 'ARRAY') {
@@ -495,9 +507,12 @@ sub _ref_prolog {
 		my @p;
 		for my $k (sort keys %$prolog ) {
 			if( defined( $prolog->{$k} ) && length( $prolog->{$k} ) ) {
-				push @p, "$k=". $prolog->{$k};
+				my $v = $prolog->{$k};
+				$k =~s/^\s+|\s+$//g;
+				push @p, "$k=$v";
 			}
 			else {
+				$k =~s/^\s+|\s+$//g;
 				unshift @p, $k;
 			}
 		}
@@ -505,9 +520,117 @@ sub _ref_prolog {
 	}
 
 	if( ref(\$prolog) eq 'SCALAR') {
-		return ['о','об=од'] if $prolog =~/^[oOоО]$/;
-		return ['с','со=с'] if $prolog =~/^[cCсС]$/;
-		return ['в'] if $prolog =~/^[BвВ]$/;
+		$prolog =~s/^\s+|\s+$//g;
+
+		# genitive - Родительный: кого? чего?
+		if( $prolog =~/^(?:
+				безо?|
+				в(?:близи|виду|доль|замен|круг|место|не|низу|нутр[иь]|переди?|роде|овнутрь|озле|округ|следствие|ыше)|
+				для|до|
+				из(?:о?|\-за|нутри|\-подо?)|
+				каса(?:ем|тельн)о|кроме|кругом|
+				мимо|
+				на(?:кануне|место|подобие|против|супротив|счет)|ниже|
+				о(?:бок|бочь|коло|крест|круг|причь|то?|тносительно)|
+				по(?:близости|верх|дле|зад[иь]|мимо|перек|се?реди(?:не)?|середь|сле|средством)|
+				пр(?:евыше|отив)|путем|
+				ради|
+				с(?:верху?|выше|ередь|зади|илами|наружи|низу|переди|ред[иь]|упротив)|
+				у
+			)$/ix
+		) {
+			$$case = 'g'; # genitive
+			return [ $prolog ];
+		}
+
+		# dative - Дательный: кому? чему?
+		if( $prolog =~ /^(?:
+				вдогон(?:ку|очку)?|вослед|вразрез|вслед|
+				ко?|
+				напере(?:кор|рез)|
+				подобно|противно|
+				соо(?:браз|тветствен)но|соразмерно
+			)$/ix
+		) {
+			$$case = 'd'; # dative
+			return [ $prolog ];
+		}
+
+		# accusative - Винительный: кого? что?
+		if( $prolog =~ /^(?:
+				(?:вы?|ис)ключая|про|сквозь|спустя|че?рез
+			)$/ix
+		) {
+			$$case = 'a'; # accusative
+			return [ $prolog ];
+		}
+
+		# instrumental - Творительный: кем? чем?
+		if( $prolog =~ /^(?:
+				кончая|надо?|начиная|передо?|по\-[зн]ад?
+			)$/ix
+		) {
+			$$case = 'i'; # instrumental
+			return [ $prolog ];
+		}
+
+		# prepositional - Предложный: о ком? о чём?
+		if( $prolog =~ /^при$/i ) {
+			$$case = 'p'; # prepositional
+			return [ $prolog ];
+		}
+
+		# genitive (Родительный) || dative (Дательный, by default)
+		if( $prolog =~ /^согласно$/i ) {
+			$$case = 'd' if ! $$case or $$case !~/^[gd]/;
+			return [ $prolog ];
+		}
+
+		# genitive (Родительный, by default) || instrumental (Творительный)
+		if( $prolog =~ /^(?:про)?меж(?:ду)?$/i ) {
+			$$case = 'g' if ! $$case or $$case !~/^[gi]/;
+			return [ $prolog ];
+		}
+
+		# accusative (Винительный) || dative (Дательный, by default)
+		if( $prolog =~ /^благодаря$/i ) {
+			$$case = 'd' if ! $$case or $$case !~/^[ad]/;
+			return [ $prolog ];
+		}
+
+		# accusative (Винительный) || instrumental (Творительный, by default)
+		if( $prolog =~ /^(?:за|подо?)$/i ) {
+			$$case = 'i' if ! $$case or $$case !~/^[ai]/;
+			return [ $prolog ];
+		}
+
+		# accusative (Винительный) || prepositional (Предложный)
+		if( $prolog =~/^[вВB][ОоOo]?$/ ) {
+			$prolog =~tr/BOo/ВОо/;
+			$$case = 'a' if ! $$case or $$case !~/^[ap]/;
+			return [ $prolog ];
+		}
+		elsif( $prolog =~ /^[oOоО][бБ]?[oOоО]?$/ ) {
+			$$case = 'p' if ! $$case or $$case !~/^[ap]/;
+			return ['о','об=од'];
+		}
+		elsif( $prolog =~ /^на$/i ) {
+			$$case = 'a' if ! $$case or $$case !~/^[ap]/;
+			return [ $prolog ];
+		}
+
+		# accusative (Винительный) || genitive (Родительный) || instrumental (Творительный, by default)
+		if( $prolog =~/^[cCсС][ОоOo]?$/ ) {
+			$$case = 'i' if ! $$case or $$case !~/^[agi]/;
+			return ['с','со=ст'];
+		}
+
+		# accusative (Винительный) || dative (Дательный, by default) || prepositional (Предложный)
+		if( $prolog =~ /^по$/i ) {
+			$$case = 'd' if ! $$case or $$case !~/^[adp]/;
+			return [ $prolog ];
+		}
+
 	}
 
 	return [ ];
@@ -527,29 +650,46 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => 'рубл',
 					'ends' => {
-							'n' => ['ей','ь', ('я') x 3, ('ей') x 3 ],	# nominative - Именительный
-							'g' => ['ей','я', ('ей') x 6 ],	# genitive - Родительный
-							'd' => ['ям','ю', ('ям') x 4, ('ей') x 2 ],	# dative - Дательный
-							'a' => ['ей','ь', ('я') x 3, ('ей') x 3 ],	# accusative - Винительный
-							'i' => ['ей','ём', ('ями') x 4, ('ей') x 2 ],	# instrumental - Творительный
-							'p' => ['ей','е', ('ях') x 4, ('ей') x 2 ],	# prepositional - Предложный
+							'n' => ['ей','ь', ('я') x 3, ('ей') x 2 ],	# nominative - Именительный
+							'g' => ['ей','я', ('ей') x 5 ],	# genitive - Родительный
+							'd' => ['ям','ю', ('ям') x 4, 'ей' ],	# dative - Дательный
+							'a' => ['ей','ь', ('я') x 3, ('ей') x 2 ],	# accusative - Винительный
+							'i' => ['ей','ём', ('ями') x 4, 'ей' ],	# instrumental - Творительный
+							'p' => ['ей','е', ('ях') x 4, 'ей' ],	# prepositional - Предложный
 					},
 					'plural' => {'n'=>'и', 'g'=>'ей', 'd'=>'ям', 'a'=>'и', 'i'=>'ями', 'p'=>'ях'},
 				);
 
 		}
-		elsif( $epilog =~/^(?:rub|\-643)$/) { # Российская копейка
+		elsif( $epilog =~/^(?:BYR|974)$/) { # Белорусский рубль
+			%eRef = (
+					'object' => 'inanimate',
+					'gender' => 'masculine',
+					'root' => ['белорусск','рубл'],
+					'ends' => {
+							'n' => [['их','ий', ('их') x 5 ], ['ей','ь', ('я') x 3, ('ей') x 2 ]],	# nominative - Именительный
+							'g' => [['их','ого', ('их') x 5 ], ['ей','я', ('ей') x 5 ]],	# genitive - Родительный
+							'd' => [['им','ому', ('им') x 4, 'их'], ['ям','ю', ('ям') x 4, 'ей' ]],	# dative - Дательный
+							'a' => [['их','ий', ('их') x 5 ], ['ей','ь', ('я') x 3, ('ей') x 2 ]],	# accusative - Винительный
+							'i' => [['их','им', ('ими') x 4, 'их'], ['ей','ём', ('ями') x 4, 'ей']],	# instrumental - Творительный
+							'p' => [['их','ом', ('их') x 5], ['ей','е', ('ях') x 4, 'ей']],	# prepositional - Предложный
+					},
+					'plural' => {'n'=>['ие','и'], 'g'=>['их','ей'], 'd'=>['им','ям'], 'a'=>['ие','и'], 'i'=>['ими','ями'], 'p'=>['их','ях']},
+				);
+
+		}
+		elsif( $epilog =~/^(?:rub|\-643|byr|\-974)$/) { # Российская | Белорусская копейка
 			%eRef = (
 					'object' => 'inanimate',
 					'gender' => 'feminine',
 					'root' => 'копе',
 					'ends' => {
-							'n' => ['ек','йка', ('йки') x 3, ('ек') x 3 ],
-							'g' => ['ек','йки', ('ек') x 6 ],
-							'd' => ['ек','йке', ('йкам') x 4, ('ек') x 2 ],
-							'a' => ['ек','йку', ('йки') x 3, ('ек') x 3 ],
-							'i' => ['ек','йкой', ('йками') x 4, ('ек') x 2 ],
-							'p' => ['ек','йке', ('йках') x 4, ('ек') x 2 ],
+							'n' => ['ек','йка', ('йки') x 3, ('ек') x 2 ],
+							'g' => ['ек','йки', ('ек') x 5 ],
+							'd' => ['ек','йке', ('йкам') x 4, 'ек' ],
+							'a' => ['ек','йку', ('йки') x 3, ('ек') x 2 ],
+							'i' => ['ек','йкой', ('йками') x 4, 'ек' ],
+							'p' => ['ек','йке', ('йках') x 4, 'ек' ],
 					},
 					'plural' => {'n'=>'йки', 'g'=>'ек', 'd'=>'йкам', 'a'=>'йки', 'i'=>'йками', 'p'=>'йках'},
 				);
@@ -561,12 +701,12 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => 'доллар',
 					'ends' => {
-							'n' => ['ов','', ('а') x 3, ('ов') x 3 ],
-							'g' => ['ов','а', ('ов') x 6 ],
-							'd' => ['ам','у', ('ам') x 4, ('ов') x 2 ],
-							'a' => ['ов','', ('а') x 3, ('ов') x 3 ],
-							'i' => ['ов','ом', ('ами') x 4, ('ов') x 2 ],
-							'p' => ['ах','е', ('ах') x 6 ],
+							'n' => ['ов','', ('а') x 3, ('ов') x 2 ],
+							'g' => ['ов','а', ('ов') x 5 ],
+							'd' => ['ам','у', ('ам') x 4, 'ов' ],
+							'a' => ['ов','', ('а') x 3, ('ов') x 2 ],
+							'i' => ['ов','ом', ('ами') x 4, 'ов' ],
+							'p' => ['ов','е', ('ах') x 4, 'ов' ],
 					},
 					'plural' => {'n'=>'ы', 'g'=>'ов', 'd'=>'ам', 'a'=>'ы', 'i'=>'ами', 'p'=>'ах'},
 				);
@@ -578,12 +718,12 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => 'цент',
 					'ends' => {
-							'n' => ['ов','', ('а') x 3, ('ов') x 3 ],
-							'g' => ['ов','а', ('ов') x 6 ],
-							'd' => ['ам','у', ('ам') x 6 ],
-							'a' => ['ов','', ('а') x 3, ('ов') x 3 ],
-							'i' => ['ами','ом', ('ами') x 6 ],
-							'p' => ['ах','е', ('ах') x 6 ],
+							'n' => ['ов','', ('а') x 3, ('ов') x 2 ],
+							'g' => ['ов','а', ('ов') x 5 ],
+							'd' => ['ам','у', ('ам') x 4, 'ов' ],
+							'a' => ['ов','', ('а') x 3, ('ов') x 2 ],
+							'i' => ['ов','ом', ('ами') x 4, 'ов' ],
+							'p' => ['ов','е', ('ах') x 4, 'ов' ],
 					},
 					'plural' => {'n'=>'ы', 'g'=>'ов', 'd'=>'ам', 'a'=>'ы', 'i'=>'ами', 'p'=>'ах'},
 				);
@@ -595,12 +735,12 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => 'юан',
 					'ends' => {
-							'n' => ['ей','ь', ('я') x 3, ('ей') x 3 ],
-							'g' => ['ей','я', ('ей') x 6 ],
-							'd' => ['ям','ю', ('ям') x 6 ],
-							'a' => ['ей','ь', ('я') x 3, ('ей') x 3 ],
-							'i' => ['ей','ем', ('ями') x 4, ('ей') x 2 ],
-							'p' => ['ях','е', ('ях') x 4, ('ей') x 2 ],
+							'n' => ['ей','ь', ('я') x 3, ('ей') x 2 ],
+							'g' => ['ей','я', ('ей') x 5 ],
+							'd' => ['ям','ю', ('ям') x 5 ],
+							'a' => ['ей','ь', ('я') x 3, ('ей') x 2 ],
+							'i' => ['ей','ем', ('ями') x 4, 'ей' ],
+							'p' => ['ях','е', ('ях') x 4, 'ей' ],
 					},
 					'plural' => {'n'=>'и', 'g'=>'ей', 'd'=>'ям', 'a'=>'и', 'i'=>'ями', 'p'=>'ях'},
 				);
@@ -615,12 +755,12 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => '',
 					'ends' => {
-							'n' => ['лет','год', ('года') x 3, ('лет') x 3 ],
-							'g' => ['лет','года', ('лет') x 6 ],
-							'd' => ['лет','году', ('годам') x 4, ('лет') x 2 ],
-							'a' => ['лет','год', ('года') x 3, ('лет') x 3 ],
-							'i' => ['лет','годом', ('годами') x 4, ('лет') x 2 ],
-							'p' => ['лет','годе', ('годах') x 4, ('лет') x 2 ],
+							'n' => ['лет','год', ('года') x 3, ('лет') x 2 ],
+							'g' => ['лет','года', ('лет') x 5 ],
+							'd' => ['лет','году', ('годам') x 4, 'лет' ],
+							'a' => ['лет','год', ('года') x 3, ('лет') x 2 ],
+							'i' => ['лет','годом', ('годами') x 4, 'лет' ],
+							'p' => ['лет','годе', ('годах') x 4, 'лет' ],
 					},
 					'plural' => {'n'=>'годы', 'g'=>'лет', 'd'=>'годам', 'a'=>'годы', 'i'=>'годами', 'p'=>'годах'},
 				);
@@ -632,12 +772,12 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => 'месяц',
 					'ends' => {
-							'n' => ['ев','', ('а') x 3, ('ев') x 3 ],
-							'g' => ['ев','а', ('ев') x 6 ],
-							'd' => ['ам','у', ('ам') x 6 ],
-							'a' => ['ев','', ('а') x 3, ('ев') x 3 ],
-							'i' => ['ев','ем', ('ами') x 4, ('ев') x 2 ],
-							'p' => ['ев','е', ('ах') x 4, ('ев') x 2 ],
+							'n' => ['ев','', ('а') x 3, ('ев') x 2 ],
+							'g' => ['ев','а', ('ев') x 5 ],
+							'd' => ['ам','у', ('ам') x 5 ],
+							'a' => ['ев','', ('а') x 3, ('ев') x 2 ],
+							'i' => ['ев','ем', ('ами') x 4, 'ев'],
+							'p' => ['ев','е', ('ах') x 4, 'ев'],
 					},
 					'plural' => {'n'=>'ы', 'g'=>'ев', 'd'=>'ам', 'a'=>'ы', 'i'=>'ами', 'p'=>'ах'},
 				);
@@ -649,12 +789,12 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => 'д',
 					'ends' => {
-							'n' => ['ней','ень', ('ня') x 3, ('ней') x 3 ],
-							'g' => ['ней','ня', ('ней') x 6 ],
-							'd' => ['ням','ню', ('ням') x 6 ],
-							'a' => ['ней','ень', ('ня') x 3, ('ней') x 3 ],
-							'i' => ['ней','нём', ('нями') x 4, ('ней') x 2 ],
-							'p' => ['ней','не', ('нях') x 4, ('ней') x 2 ],
+							'n' => ['ней','ень', ('ня') x 3, ('ней') x 2 ],
+							'g' => ['ней','ня', ('ней') x 5 ],
+							'd' => ['ням','ню', ('ням') x 5 ],
+							'a' => ['ней','ень', ('ня') x 3, ('ней') x 2 ],
+							'i' => ['ней','нём', ('нями') x 4, 'ней'],
+							'p' => ['ней','не', ('нях') x 4, 'ней'],
 					},
 					'plural' => {'n'=>'ни', 'g'=>'ней', 'd'=>'ням', 'a'=>'ни', 'i'=>'нями', 'p'=>'нях'},
 				);
@@ -666,12 +806,12 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => 'час',
 					'ends' => {
-							'n' => ['ов','', ('а') x 3, ('ов') x 3 ],
-							'g' => ['ов','а', ('ов') x 6 ],
-							'd' => ['ам','у', ('ам') x 6 ],
-							'a' => ['ов','', ('а') x 3, ('ов') x 3 ],
-							'i' => ['ов','ом', ('ами') x 4, ('ов') x 2 ],
-							'p' => ['ов','е', ('ах') x 4, ('ов') x 2 ],
+							'n' => ['ов','', ('а') x 3, ('ов') x 2 ],
+							'g' => ['ов','а', ('ов') x 5 ],
+							'd' => ['ам','у', ('ам') x 5 ],
+							'a' => ['ов','', ('а') x 3, ('ов') x 2 ],
+							'i' => ['ов','ом', ('ами') x 4, 'ов'],
+							'p' => ['ов','е', ('ах') x 4, 'ов'],
 					},
 					'plural' => {'n'=>'ы', 'g'=>'ов', 'd'=>'ам', 'a'=>'ы', 'i'=>'ами', 'p'=>'ах'},
 				);
@@ -683,12 +823,12 @@ sub _ref_epilog {
 					'gender' => 'feminine',
 					'root' => 'минут',
 					'ends' => {
-							'n' => ['','а', ('ы') x 3, ('') x 3 ],
-							'g' => ['','ы', ('') x 6 ],
-							'd' => ['','е', ('ам') x 4, ('') x 2 ],
-							'a' => ['','у', ('ы') x 3, ('') x 3 ],
-							'i' => ['','ой', ('ами') x 4, ('') x 2 ],
-							'p' => ['','е', ('ах') x 4, ('') x 2 ],
+							'n' => ['','а', ('ы') x 3, ('') x 2 ],
+							'g' => ['','ы', ('') x 5 ],
+							'd' => ['','е', ('ам') x 4, ''],
+							'a' => ['','у', ('ы') x 3, ('') x 2 ],
+							'i' => ['','ой', ('ами') x 4, ''],
+							'p' => ['','е', ('ах') x 4, ''],
 					},
 					'plural' => {'n'=>'ы', 'g'=>'', 'd'=>'ам', 'a'=>'ы', 'i'=>'ами', 'p'=>'ах'},
 				);
@@ -700,12 +840,12 @@ sub _ref_epilog {
 					'gender' => 'feminine',
 					'root' => 'секунд',
 					'ends' => {
-							'n' => ['','а', ('ы') x 3, ('') x 3 ],
-							'g' => ['','ы', ('') x 6 ],
-							'd' => ['','е', ('ам') x 4, ('') x 2 ],
-							'a' => ['','у', ('ы') x 3, ('') x 3 ],
-							'i' => ['','ой', ('ами') x 4, ('') x 2 ],
-							'p' => ['','е', ('ах') x 4, ('') x 2 ],
+							'n' => ['','а', ('ы') x 3, ('') x 2 ],
+							'g' => ['','ы', ('') x 5 ],
+							'd' => ['','е', ('ам') x 4, ''],
+							'a' => ['','у', ('ы') x 3, ('') x 2 ],
+							'i' => ['','ой', ('ами') x 4, ''],
+							'p' => ['','е', ('ах') x 4, ''],
 					},
 					'plural' => {'n'=>'ы', 'g'=>'', 'd'=>'ам', 'a'=>'ы', 'i'=>'ами', 'p'=>'ах'},
 				);
@@ -717,12 +857,12 @@ sub _ref_epilog {
 					'gender' => 'masculine',
 					'root' => 'метр',
 					'ends' => {
-							'n' => ['ов','', ('а') x 3, ('ов') x 3 ],
-							'g' => ['ов','а', ('ов') x 6 ],
-							'd' => ['ов','у', ('ам') x 4, ('ов') x 2 ],
-							'a' => ['ов','', ('а') x 3, ('ов') x 3 ],
-							'i' => ['ов','ом', ('ами') x 4, ('ов') x 2 ],
-							'p' => ['ов','е', ('ах') x 4, ('ов') x 2 ],
+							'n' => ['ов','', ('а') x 3, ('ов') x 2 ],
+							'g' => ['ов','а', ('ов') x 5 ],
+							'd' => ['ов','у', ('ам') x 4, 'ов'],
+							'a' => ['ов','', ('а') x 3, ('ов') x 2 ],
+							'i' => ['ов','ом', ('ами') x 4, 'ов'],
+							'p' => ['ов','е', ('ах') x 4, 'ов'],
 					},
 					'plural' => {'n'=>'ы', 'g'=>'ов', 'd'=>'ам', 'a'=>'ы', 'i'=>'ами', 'p'=>'ах'},
 				);
@@ -734,12 +874,12 @@ sub _ref_epilog {
 					'gender' => 'feminine',
 					'root' => 'печат',
 					'ends' => {
-							'n' => ['ей','ь', ('и') x 3, ('ей') x 3 ],
-							'g' => ['ей','и', ('ей') x 6 ],
-							'd' => ['ей','и', ('ям') x 4, ('ей') x 2 ],
-							'a' => ['ей','ь', ('и') x 3, ('ей') x 3 ],
-							'i' => ['ей','ью', ('ями') x 4, ('ей') x 2 ],
-							'p' => ['ей','и', ('ях') x 4, ('ей') x 2 ],
+							'n' => ['ей','ь', ('и') x 3, ('ей') x 2 ],
+							'g' => ['ей','и', ('ей') x 5 ],
+							'd' => ['ей','и', ('ям') x 4, 'ей'],
+							'a' => ['ей','ь', ('и') x 3, ('ей') x 2 ],
+							'i' => ['ей','ью', ('ями') x 4, 'ей'],
+							'p' => ['ей','и', ('ях') x 4, 'ей'],
 					},
 					'plural' => {'n'=>'и', 'g'=>'ей', 'd'=>'ям', 'a'=>'и', 'i'=>'ями', 'p'=>'ях'},
 				);
@@ -750,12 +890,29 @@ sub _ref_epilog {
 			$$object = $eRef{'object'};
 			$$gender = $eRef{'gender'};
 
-			# To fix for plural '1' only 
-			$eRef{'ends'}{ $case }[1] = $eRef{'plural'}{ $case } if $multi =~/^plural/i;
+			# To fix for plural '1' only
+			if( $multi =~/^plural/i ) {
+				if( ref( \$eRef{'plural'}{ $case } ) eq 'SCALAR') {
+					$eRef{'ends'}{ $case }[1] = $eRef{'plural'}{ $case };
+				}
+				elsif( ref( $eRef{'plural'}{ $case } ) eq 'ARRAY') {
+					my $i = 0;
+					for( @{ $eRef{'plural'}{ $case } } ) {
+						$eRef{'ends'}{ $case }[ $i++ ][1] = $_;
+					}
+				}
+			}
 
 			$epilog = {};
-			$epilog->{'root'} = $eRef{'root'};
-			$epilog->{'ends'} = $eRef{'ends'}{ $case };
+
+			if( ref( \$eRef{'root'} ) eq 'SCALAR') {
+				$epilog->{'root'} = [ $eRef{'root'} ];
+				$epilog->{'ends'}[0] = $eRef{'ends'}{ $case };
+			}
+			elsif( ref( $eRef{'root'} ) eq 'ARRAY') {
+				$epilog->{'root'} = $eRef{'root'};
+				$epilog->{'ends'} = $eRef{'ends'}{ $case };
+			}
 		}
 		else {
 			$epilog = {};
@@ -769,6 +926,7 @@ sub _ref_epilog {
 
 		$$gender = @{{'m'=>'masculine', 'f'=>'feminine', 'n'=>'neuter'}}{lc $1}
 			if exists( $epilog->{'gender'} ) && $epilog->{'gender'} =~/^\s*(m|f|n)/i;
+
 	}
 	else {
 		$epilog = {};
@@ -884,10 +1042,11 @@ Will print the result:
 
 Using reference C<'epilog'> with other default options:
 
-  say num2cardinal( 1000, 'epilog'=>'RUB');	# add Российский 'рубль', using an alphabetic currency code
+  say num2cardinal( 1000, 'epilog'=>'RUB');	# add 'рубль' (российский), using an alphabetic currency code
   say num2cardinal( 1000, 'epilog'=> 643 );	# the same, using a digital currency code
-  say num2cardinal( 2000, 'epilog'=>'rub');	# add Российская 'копейка'
+  say num2cardinal( 2000, 'epilog'=>'rub');	# add 'копейка' (российская)
   say num2cardinal( 2000, 'epilog'=> -643);	# the same
+  say num2cardinal( 1000, 'epilog'=>'BYR');	# add 'белорусский рубль', using an alphabetic currency code
 
 Will print the results:
 
@@ -895,10 +1054,39 @@ Will print the results:
   одна тысяча рублей
   две тысячи копеек
   две тысячи копеек
+  одна тысяча белорусских рублей
 
 
 =item 9.
 
+Using B<custom C<'epilog'> > which is configured manually (норвежская крона) with other default options:
+
+  my $epilog = { # Nominative case (Именительный падеж)
+               'object' => 'inanimate',
+               'gender' => 'feminine',
+               'root'   => ['норвежск','крон'],
+               'ends'   => [ ['их','ая', ('их') x 5 ], ['','а', ('ы') x 3, ('') x 2] ],
+              };
+
+  say num2cardinal( $_,
+                   'epilog' => $epilog
+                  ) for 0..5, 10, 100, 1000;
+
+Will print the result:
+
+  ноль норвежских крон
+  одна норвежская крона
+  две норвежских кроны
+  три норвежских кроны
+  четыре норвежских кроны
+  пять норвежских крон
+  десять норвежских крон
+  сто норвежских крон
+  одна тысяча норвежских крон
+
+
+=item 10.
+
 Using C<'prolog'> and C<'epilog'> with other explicit and implicit options:
 
   say num2cardinal( 11,
@@ -912,8 +1100,8 @@ OR the same with details:
   my $epilog = {
                'object' => 'inanimate',
                'gender' => 'feminine',
-               'root'   => 'печат',
-               'ends'   => ['ях','и', ('ях') x 8 ],
+               'root'   => ['печат'],
+               'ends'   => [ ['ях','и', ('ях') x 5 ] ],
               };
 
   say num2cardinal( 11,
@@ -927,7 +1115,7 @@ Will print the result:
   об одиннадцати печатях
 
 
-=item 10.
+=item 11.
 
 Using C<'alt'> (alternate form) with other default options:
 
@@ -940,7 +1128,7 @@ Will print the results:
   нуль
 
 
-=item 11.
+=item 12.
 
 Using C<'ucfirst'> with other explicit and implicit options:
 
@@ -1007,12 +1195,12 @@ Lingua::RU::Numeral provides this subroutine:
 
 =head2 num2cardinal( $number [, %facultative_options ] )
 
-Convert a C<$number> (natural numbers or 0) to Russian text (string), using explicitly specified C<%facultative_options>,
-otherwise - default options. The following can be  C<%facultative_options>:
+Convert a C<$number> (natural numbers or 0) to Russian text (string), using explicitly specified global C<%facultative_options>,
+otherwise - default options. The following can be global C<%facultative_options>:
 
 =over 3
 
-=item * C<'case'> option — case of a numeral can take the following meanings (case names can be abbreviated to one first letter):
+=item * C<'case'> option — case of a numeral can take the following meanings (case names can be shorthand to one first letter):
 
 =over 6
 
@@ -1037,16 +1225,16 @@ C<'p'|'prepositional'> (I<Предложный падеж: думаю о ком?
 =back
 
 =item * C<'gender'> option — C<'m'|'masculine'> (I<Мужской род>, B<by default>) or C<'f'|'feminine'> (I<Женский род>)
-or C<'n'|'neuter'> (I<Средний род>). Gender names can be abbreviated to one first letter also.
+or C<'n'|'neuter'> (I<Средний род>). Gender names can be shorthand to one first letter also.
 
 =item * C<'multi'> option — word(s) in the plural (C<'1'|'plural'>, I<множественное число>) or C<'undef'|'0'|'singular'> (I<единственное число>, B<by default>).
 
-=item * C<'object'> option — C<'inanimate'> (I<неодушевлённый предмет>, B<by default>) or C<'animate'> (I<одушевлённый предмет>) object.
-Object names can be abbreviated to one first letter also.
+=item * C<'object'> option — C<'i'|'inanimate'> (I<неодушевлённый предмет>, B<by default>) or C<'a'|'animate'> (I<одушевлённый предмет>) object.
+Object names can be shorthand to one first letter also.
 
 =item * C<'prolog'> option — PREPosition(s) of the numeral is the REF to ARRAY,
-where 0th element ('PREP_0') — the preposition for all numerals, except for those indicated by 
-subsequent array elements ('PREP_x') as C<['PREP_0','PREP_1=REGEX',...]>.
+where 0th element ('PREP_0') — the preposition for all numerals, except for those indicated 
+by subsequent array elements ('PREP_x') as C<['PREP_0','PREP_1=REGEX',...]>.
 
 Here REGEX — the regular expression that is used to find a match at the beginning of a numeric string like:
 C<numeral =~/^REGEX/>.
@@ -1054,21 +1242,104 @@ For example:
 
   'prolog' => ['о','об=од']
 
-If C<'prolog'> is SCALAR value (e.g. C<< 'prolog'=>'о' >>) then it is used from internal B<reference list of prologs>.
-Now internal B<reference list of prologs> contains the following preconfigured phrase structures:
+If C<'prolog'> is SCALAR value (e.g. C<< 'prolog'=>'о' >>) then it is used from internal B<reference list of prologs> (inner prolog).
+In this case, the inner prolog controls B<case of the numeral> and overrides (suppresses or affirms its authority)
+the global option C<'case'>.
+
+Now internal B<reference list of prologs> contains the following preconfigured phrase structures, i.e. the inner prolog:
 
 =over 6
 
-=item * C<'o'> — equivalent to C<['о','об=од']>
+=item *
+for B<genitive> case (I<Родительный падеж>):
 
-=item * C<'c'> — equivalent to C<['с','со=с']>
+C<без>, C<безо>,
+C<вблизи>, C<ввиду>, C<вдоль>, C<взамен>, C<вкруг>, C<вместо>, C<вне>, C<внизу>, C<внутри>, C<внутрь>, C<вперед>,
+C<вследствие>, C<впереди>, C<вроде>, C<вовнутрь>, C<возле>, C<вокруг>, C<вследствие>, C<выше>,
+C<для>, C<до>,
+C<из>, C<изо>, C<из-за>, C<изнутри>, C<из-под>, C<из-подо>,
+C<касаемо>, C<касательно>, C<кроме>, C<кругом>,
+C<мимо>,
+C<накануне>, C<наместо>, C<наподобие>, C<напротив>, C<насупротив>, C<насчет>, C<ниже>,
+C<обок>, C<обочь>, C<около>, C<окрест>, C<округ>, C<опричь>, C<от>, C<ото>, C<относительно>,
+C<поблизости>, C<поверх>, C<подле>, C<позади>, C<позадь>, C<помимо>, C<поперек>, C<посереди>, C<посередине>, C<после>,
+C<посередь>, C<посреди>, C<посредине>, C<посредством>,
+C<превыше>, C<против>, C<путем>,
+C<ради>,
+C<сверх>, C<сверху>, C<свыше>, C<середь>, C<сзади>, C<силами>, C<снаружи>, C<снизу>, C<спереди>, C<среди>, C<средь>, C<супротив>,
+C<у> — equivalent to the corresponding word, specified as the only array element.
 
-=item * C<'в'> — equivalent to C<['в']>
+=item *
+for B<dative> case (I<Дательный падеж>):
+
+C<вдогон>, C<вдогонку>, C<вдогоночку>, C<вослед>, C<вразрез>, C<вслед>,
+C<к>, C<ко>, C<наперекор>, C<наперерез>, C<подобно>, C<противно>, C<сообразно>, C<соответственно>,
+C<соразмерно> — equivalent to the corresponding word, specified as the only array element.
+
+=item *
+for B<accusative> case (I<Винительный падеж>):
+
+C<включая>, C<выключая>, C<исключая>, C<про>, C<сквозь>, C<спустя>, C<через>, C<чрез> — equivalent to the corresponding word,
+specified as the only array element.
+
+=item *
+for B<instrumental> case (I<Творительный падеж>):
+
+C<кончая>, C<над>, C<надо>, C<начиная>, C<перед>, C<передо>, C<по-за>, C<по-над> — equivalent to the corresponding word,
+specified as the only array element.
+
+=item *
+for B<prepositional> case (I<Предложный падеж>):
+
+C<при> — equivalent to C<['при']>.
+
+=item *
+for B<genitive> (I<Родительный>) and B<dative> cases (I<Дательный падеж>):
+
+C<согласно> — equivalent to C<['согласно']>.
+
+=item *
+for B<genitive> (I<Родительный>) and B<instrumental> cases (I<Творительный падеж>):
+
+C<меж>, C<между>, C<промеж>, C<промежду> — equivalent to the corresponding word, specified as the only array element.
+
+=item *
+for B<accusative> (I<Винительный>) and B<instrumental> cases (I<Творительный падеж>):
+
+C<за>, C<под>, C<подо> — equivalent to the corresponding word, specified as the only array element.
+
+=item *
+for B<dative> (I<Дательный падеж>) and B<accusative> cases (I<Винительный падеж>):
+
+C<благодаря> — equivalent to C<['благодаря']>.
+
+=item *
+for B<accusative> (I<Винительный>) and B<prepositional> cases (I<Предложный падеж>):
+
+на — equivalent to C<['на']>.
+
+C<o>, C<oб>, C<oбо> — equivalent to C<['о','об=од']>.
+
+C<в>, C<во> — equivalent to C<['в']> and C<['во']>, respectively.
+
+=item *
+for B<genitive> (I<Родительный>), B<accusative> (I<Винительный>), and B<instrumental> cases (I<Творительный падеж>):
+
+C<c>, C<со> — equivalent to C<['с','со=ст']>.
+
+=item *
+for B<dative> (I<Дательный>), B<accusative> (I<Винительный>), and B<prepositional> cases (I<Предложный падеж>):
+
+C<по> — equivalent to C<['по']>.
 
 =back
 
-=item * C<'epilog'> option — final word (phrase) of the numeral is SCALAR value (name) from internal B<reference list of epilogs>
-or the REF to HASH to create a B<custom> phrase.
+B<WARNING!> Once again, the inner prolog controls B<case of the numeral> and overrides (suppresses or affirms its authority)
+the global option C<'case'>.
+
+
+=item * C<'epilog'> option — final word(s) (phrase(s)) of the numeral is SCALAR value (name) from internal B<reference list of epilogs>
+or the REF to complex HASH to create a B<custom> phrase(s).
 
 Internal B<reference list of epilogs> contains the following preconfigured phrase structures:
 
@@ -1082,6 +1353,9 @@ C<'rub'|-643> — for Russian kopek (I<Российская копейка>).
 =item * C<'USD'|840> — for United States dollar;
 C<'usd'|-840> — for US cent.
 
+=item * C<'BYR'|974> — for Belarusian ruble (I<Белорусский рубль>);
+C<'byr'|-974> — for kopek (I<копейка>).
+
 =item * Time measures:  C<'year'> — for I<год>, I<лет>;
 C<'month'> — I<месяц>;
 C<'day'> — I<день>;
@@ -1095,32 +1369,32 @@ C<'sec.'> — second (I<секунда>).
 
 =back
 
-The B<custom C<'epilog'> > presents the REF to HASH with complex structure of SCALAR values and ARRAY.
-For example, I<год>, I<лет> (abbreviation equivalent is C<$epilog = 'year'> for C<'nominative'> case (Именительного падежа):
+The B<custom C<'epilog'> > presents the REF to HASH with complex structure of SCALAR values and ARRAYs.
+For example, I<год>, I<лет> (shorthand equivalent is C<$epilog = 'year'> for C<'nominative'> case (Именительного падежа):
 
   my $epilog = {
                'object' => 'inanimate', # неодушевлённый предмет
                'gender' => 'masculine', # Мужской род
-               'root'   => '',
-               'ends'   => ['лет','год', ('года') x 3, ('лет') x 3 ],
+               'root'   => [''],
+               'ends'   => [['лет','год', ('года') x 3, ('лет') x 2 ]],
               };
 
-or for I<секунда> (abbreviation equivalent is C<$epilog = 'sec.'> for C<'accusative'> case (Винительного падежа)):
+or for I<секунда> (shorthand equivalent is C<$epilog = 'sec.'> for C<'accusative'> case (Винительного падежа)):
 
   my $epilog = {
                'object' => 'inanimate',
                'gender' => 'feminine', # Женский род
-               'root'   => 'секунд',
-               'ends'   => ['','у', ('ы') x 3, ('') x 3 ],
+               'root'   => ['секунд'],
+               'ends'   => [['','у', ('ы') x 3, ('') x 2 ]],
               };
 
-or for I<рубль> (abbreviation equivalent is C<$epilog = 'RUB'> for C<'genitive'> case (Родительного падежа)):
+or for I<рубль> (shorthand equivalent is C<$epilog = 'RUB'> for C<'dative'> case (Дательного падежа)):
 
   my $epilog = {
                'object' => 'inanimate',
                'gender' => 'masculine', # Мужской род
-               'root'   => 'рубл',
-               'ends'   => ['ям','ю', ('ям') x 4, ('ей') x 2 ],
+               'root'   => ['рубл'],
+               'ends'   => [['ям','ю', ('ям') x 4, 'ей']],
               };
 
 Here are:
@@ -1133,8 +1407,9 @@ B<WARNING!> These options take precedence over the global options(C<'object'>, C
 =item * C<'root'> option means the C<'epilog'> root, i.e. common invariable part of the custom C<'epilog'> words
 (or its absence in the form of "").
 
-=item * C<'ends'> option means the C<'epilog'> word endings for numerals: 0, 1, 2, 3, 4, 5,
-1000 (thousand), 1_000_000 (million) respectively.
+=item * C<'ends'> option means the C<'epilog'> word endings for seven (7) numerals: 0, 1, 2, 3, 4, 5, 1000 (thousand) respectively.
+In fact, the seventh ending in each subarray also corresponds to a million (1_000_000), a billion (1_000_000_000),
+and a trillion (1_000_000_000_000).
 
 =back
 
@@ -1142,7 +1417,7 @@ B<WARNING!> These options take precedence over the global options(C<'object'>, C
 =item * C<'alt'> option — REF to HASH of alternative word forms.
 Currently available for zero only (I<ноль>, by default; C<< 'alt' => {0=>1} >>, to obtain I<нуль>).
 
-=item * C<'ucfirst'> option — Returns the numeral with the first character capitalized (upper case).
+=item * C<'ucfirst'> option — Returns the numeral with the first character capitalized (Upper case).
 B<WARNING!> does not affect the C<'prolog'>.
 
 =back
@@ -1173,6 +1448,12 @@ Institut für Slavistik der Universität, -1985. — 514 p. ISSN 0258-6819.
 Мельчук, И.А. Поверхностный синтаксис русских числовых выражений. -Wien: Volume 16 of Wiener slawistischer Almanach,
 Institut für Slavistik der Universität, -1985. — 514 c. ISSN 0258-6819.
 
+Сичинава, Д.В. Числительные // Материалы к корпусной грамматике русского языка. Выпуск III : Части речи и лексико-грамматические классы.
+СПб.: Нестор-История, 2018. С. 193–257.
+
+Сичинава, Д.В. Предлоги // Материалы к корпусной грамматике русского языка. Выпуск III : Части речи и лексико-грамматические классы.
+СПб.: Нестор-История, 2018. С. 329–373.
+
 L<Lingua::RU::Number> is a Perl module that offers some similar functionality.
 
 =head1 AUTHOR
